/**
 * Mapper class is a helper to create projection mapping with Processing.
 * It allows to align textures on a projection to a real life object
 * by dragging corners of each texture. 
 * Use SPACE to toggle editing mode on and off.
 * Use TAB to switch between active texture (you can drag corners 
 * of an active texture to adjust the projection).
 * Use ARROWS UP DOWN LEFT RIGHT to move currently selected corner. 
 * You can do it with the mouse also.
 * Use 's' key to save current configuration (mappings).
 * Use 'l' key to load previously saved configuration.
 * Mappings are stored in a json in data/mappings.json
 *
 * The class is configured to use 3 square textures with 4 corners by default.
 * Feel free to experiment.
 *
 * @author Michał Urbański (tajnyprojekt.com)
 * MIT license
 * 2020
 */
class Mapper {
  final String filename = "mappings.json";
  int activeFigure = 0;
  int activeCorner = 0;
  boolean isEditing = true;
  boolean isDragging = false;  
  
  static final int FIG_NUM = 3; // figures count - update this values when changing corners array below
  static final int COR_NUM = 4; // corners count in each figure - update this values when changing corners array below
  
  // set default corners positions
  PVector corners[][] = new PVector[][] {
    {
      new PVector(612, 120),
      new PVector(412, 220),
      new PVector(612, 384),
      new PVector(812, 220)
    },
    {
      new PVector(412, 220),
      new PVector(412, 580),
      new PVector(612, 620),
      new PVector(612, 384)
    },
    {
      new PVector(612, 384),
      new PVector(612, 620),
      new PVector(812, 580),
      new PVector(812, 220)    
    },
  };
  
  PVector textureMapping[] = new PVector[] {
    new PVector(0, 0),
    new PVector(0, 1),
    new PVector(1, 1),
    new PVector(1, 0)
  };
  
  Mapper() {
    File f = new File(dataPath(filename));
    if (f.exists()) {
      println("found saved mappings");
      loadMappings();
    }
  }
  
  void drawContent(PGraphics tex[]) {
    pushStyle();
    for (int f = 0; f < FIG_NUM; f++) {
      beginShape(QUAD);
      noStroke();  
      fill(f * 50 + 50); 
      texture(tex[f]);
      for (int c = 0; c < COR_NUM; c++) {
        PVector corner = corners[f][c];
        PVector texMap = textureMapping[c];
        vertex(corner.x, corner.y, texMap.x, texMap.y);
      }
      endShape();
    }
    popStyle();
  }
  
  void drawGUI() {
    if (isEditing) {
      pushStyle();
      strokeWeight(2);
      noFill();
      drawCrosshair();
      dragCorner();
      for (int f = 0; f < FIG_NUM; f++) {
        stroke(200, 200, 200); 
        if (f != activeFigure) {
          drawFigure(f);
        }
      }
      stroke(255, 100, 100);
      fill(255, 100, 100, 150);
      drawFigure(activeFigure);
      drawCornerCircles();
      popStyle();
    }
  }
  
  void drawCrosshair() {
    pushStyle();
    stroke(255);
    line(0, mouseY, width, mouseY);
    line(mouseX, 0, mouseX, height);
    popStyle();
  }
  
  void drawFigure(int f) {
    beginShape();
    for (int c = 0; c < COR_NUM; c++) {
      PVector corner = corners[f][c];
      vertex(corner.x, corner.y);
    }
    endShape(CLOSE);
  }
  
  void drawCornerCircles() {
    for (int c = 0; c < COR_NUM; c++) {
      PVector corner = corners[activeFigure][c];
      noFill();
      float dist = corner.dist(new PVector(mouseX, mouseY));
      if (c == activeCorner) {
        ellipse(corner.x, corner.y, 20, 20);
      }
      else if (dist < 15) {
        ellipse(corner.x, corner.y, 15, 15);
      }
    }
  }
  
  void dragCorner() {
    PVector aC = corners[activeFigure][activeCorner];
    if (isDragging) {
      aC.x = mouseX;
      aC.y = mouseY;
    }
  }
  
  void saveMappings() {
    JSONArray mappings = new JSONArray();
    for (int f = 0; f < FIG_NUM; f++) {
      JSONArray arr = new JSONArray();
      for (int c = 0; c < COR_NUM; c++) {
        PVector corner = corners[f][c];
        JSONObject obj = new JSONObject();
        obj.setFloat("x", corner.x);
        obj.setFloat("y", corner.y);
        arr.setJSONObject(c, obj);
      }
      mappings.setJSONArray(f, arr);
    }
    saveJSONArray(mappings, dataPath(filename));
    println("mappings saved to: " + filename);
  }
  
  void loadMappings() {
    JSONArray mappings = loadJSONArray(filename);
    for (int f = 0; f < FIG_NUM; f++) {
      JSONArray arr = mappings.getJSONArray(f);
      for (int c = 0; c < COR_NUM; c++) {
        JSONObject obj = arr.getJSONObject(c);
        corners[f][c] = new PVector(obj.getFloat("x", 100), obj.getFloat("y"));
      }
    }
    println("loaded mapping from: " + filename);
  }
  
  void onKeyPressed() {
    if (isEditing) {
      PVector aC = corners[activeFigure][activeCorner];
      if (key == CODED) {
        if (keyCode == LEFT) {
          aC.x--;
        }
        if (keyCode == RIGHT) {
         aC.x++; 
        }
        if (keyCode == UP) {
          aC.y--;
        }
        if (keyCode == DOWN) {
          aC.y++;
        }
      }
    }
  }
  
  void onKeyReleased() {
    if (key == ' ') {
      isEditing = !isEditing;
      if (isEditing) {
        println("enabled mapping editor");
      }
      else {
        println("disabled mapping editor");
      }
    }
    if (isEditing) {
      if (key == TAB) {
        activeFigure = ++activeFigure % FIG_NUM;
      }
      if (key == 's') {
        saveMappings();
      }
      if (key == 'l') {
        loadMappings();
      }
    }
  }

  void onMousePressed() {
    if (isEditing) {
      for (int c = 0; c < COR_NUM; c++) {
        PVector corner = corners[activeFigure][c];
        float dist = corner.dist(new PVector(mouseX, mouseY));
        if (dist < 15) {
          activeCorner = c;
          isDragging = true;
        }
      }
    }
  }
  
  void onMouseReleased() {
    if (isEditing) {
      isDragging = false;
    }
  }
  
}